@@ -448,7 +448,7 @@ mod tests {
     fn must_panic_on_out_of_bounds_range() {
         test_everything::<Compact>(
             Constructor::New,
-            vec![Action::Slice(TestBounds::Range(0, 13764126361151078400))],
+            vec![Action::Slice(TestBounds::Range(0, usize::MAX - 1))],
         );
     }
 
@@ -525,7 +525,30 @@ mod tests {
     }
 
     #[test]
-    fn string_layout_consistency_check() {
-        crate::validate();
+    fn check_alignment() {
+        use crate::boxed::BoxedString;
+        use crate::inline::InlineString;
+        use crate::marker_byte::Discriminant;
+
+        let inline = InlineString::new();
+        let inline_ptr: *const InlineString = &inline;
+        let boxed_ptr: *const BoxedString = inline_ptr.cast();
+        #[allow(unsafe_code)]
+        let discriminant =
+            Discriminant::from_bit(BoxedString::check_alignment(unsafe { &*boxed_ptr }));
+        assert_eq!(Discriminant::Inline, discriminant);
+
+        let boxed = BoxedString::from_str(32, "welp");
+        let discriminant = Discriminant::from_bit(BoxedString::check_alignment(&boxed));
+        assert_eq!(Discriminant::Boxed, discriminant);
+
+        let mut s = SmartString::<Compact>::new();
+        assert_eq!(Discriminant::Inline, s.discriminant());
+        let big_str = "1234567890123456789012345678901234567890";
+        assert!(big_str.len() > MAX_INLINE);
+        s.push_str(big_str);
+        assert_eq!(Discriminant::Boxed, s.discriminant());
+        s.clear();
+        assert_eq!(Discriminant::Inline, s.discriminant());
     }
 }
